from datetime import datetime

from .base import PersistenceFacade

from ..models import multi_query_serialize, Query, get_json_serializer

class SpinHubPersistenceFacade(PersistenceFacade):
    def __init__(self, spinhub):
        self.spinhub = spinhub

    @property
    def session(self):
        return self.spinhub.session

    @property
    def tests_endpoint(self):
        return '/tests/{}'.format(self.spinhub.default_org['key'])

    def create(self, records):
        serialize = get_json_serializer().serialize
        records = [serialize(tr) for tr in list(records)]
        return self.session.post(self.tests_endpoint, json=dict(tests=records))
        
    def retrieve(self, test_selector=None, feature_selector=None):
        queries = multi_query_serialize(test=test_selector, feat=feature_selector)
        return self.session.get(self.tests_endpoint, params=queries)
        
    def update(self, records):
        raise NotImplementedError()
    
    def delete(self, match_query):
        return self.session.delete(self.tests_endpoint, params=match_query)



import json
import jwt
import requests

from json.decoder import JSONDecodeError

from simple_memory_cache import GLOBAL_CACHE

from ..errors import AuthUnauthorized


class Auth0Provider(object):
    def __init__(self, domain, audience, client_id, jwks_url, user_info_url):
        self.domain = domain
        self.audience = audience
        self.client_id = client_id
        self.jwks_url = jwks_url
        self.user_info_url = user_info_url

        self.cached_jwks = GLOBAL_CACHE.MemoryCachedVar('auth_jwks')
        self.cached_jwks.on_first_access(self._retrieve_jwks)

    def _retrieve_jwks(self):
        return requests.get(self.jwks_url).json()
    
    def get_user_info(self, access_token):
        return requests.get(self.user_info_url, headers={'Authorization': 'Bearer ' + access_token}).json()

    def authenticate(self, username, password, scopes=[]):
        return self._execute_request('/oauth/token',
            grant_type = 'password',
            username = username,
            password = password,
            audience = self.audience,
            client_id = self.client_id,
            scope = 'offline_access ' + ' '.join(scopes) # Special scope to receive a refresh token
        )
    
    def _execute_request(self, url_path, **data):
        URL = 'https://' + self.domain  + url_path
        resp = requests.post(URL, data = data)
        content = self._handle_errors_in_content(resp)
        return content
    
    def _handle_errors_in_content(self, resp):
        try:
            content = resp.json()
        except JSONDecodeError:
            content = {}
            
        if resp.status_code >= 400:
            desc = content.get('error_description', None)
            if desc:
                raise AuthUnauthorized(desc)
            else:
                raise AuthUnauthorized(str(content))
    
        return content
    
    def refresh_access_token(self, refresh_token):
        return self._execute_request('/oauth/token',
            grant_type = 'refresh_token',
            client_id = self.client_id,
            refresh_token = refresh_token
        )

    def revoke_refresh_token(self, refresh_token):
        self._execute_request('/oauth/revoke',
            client_id = self.client_id,
            token = refresh_token
        )
    
    def get_jwks_key(self, access_token):
        jwks = self.cached_jwks.get()
        public_keys = {}
        for jwk in jwks['keys']:
            kid = jwk['kid']
            public_keys[kid] = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk))
            
        kid = jwt.get_unverified_header(access_token)['kid']
        key = public_keys[kid]
        return key

    def jwt_verify(self, access_token):
        try:
            payload = self.decode(access_token)
        except jwt.PyJWTError:
            # invalidate jwks cache and try ONCE more.
            self.cached_jwks.invalidate()
            payload = self.decode(access_token)
        return payload
    
    def decode(self, access_token):
        key = self.get_jwks_key(access_token)
        return jwt.decode(access_token, key=key, algorithms=['RS256'], audience=[self.client_id, self.audience])
    

class SpinHubAuthProvider(Auth0Provider):
    def __init__(self):
        domain = "spinhub.auth0.com"
        super(SpinHubAuthProvider, self).__init__(
            domain = domain,
            audience = "https://app.spinhub.ca/api",
            client_id = "fYeWIlWMD99S6eJi86pKN7oa2ll2k7lE",
            jwks_url = 'https://' + domain + '/.well-known/jwks.json',
            user_info_url = 'https://' + domain + '/userinfo'
        )
    
    



import re

class MongoQuery():
    def __init__(self, model_query):
        self.orig_query = model_query

    def build(self):
        query = {}

        if self.orig_query is None:
            return query

        for key, value in self.orig_query.value_equals.items():
            query[key] = value
        
        for key, value in self.orig_query.list_contains.items():
            raise NotImplementedError()

        for key, list_of_values in self.orig_query.value_equals_one_of.items():
            query[key] = q_any_in_list(list_of_values)
        
        return query
    
def q_any_in_list(_list):
    return {'$in': _list}


import logging

LEVEL = logging.DEBUG
CORE_LOGGER = logging.getLogger('sptp')
FORMAT = '%(name)-30s:%(levelname)-5s: %(message)s'

def _logger(name):
    return CORE_LOGGER.getChild(name)

def setup_logging():
    CORE_LOGGER.setLevel(LEVEL)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter(FORMAT))
    
    CORE_LOGGER.addHandler(stream_handler)


import pkgutil
import socket
from contextlib import closing

version_file_content = pkgutil.get_data('spintop.compat', 'VERSION' )
VERSION = version_file_content.decode().strip()
    
def format_exception(e):
    return e.__class__.__name__ + ': ' + str(e)
    

def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


from .base import (
    TransformBuilder, 
    Transformer, 
    join_transforms, 
    transformer
)

from .dut_op_log import DutOp, DutOpLog, DutOpLogTransformer

from .misc import (
    ForceTestbenchNameTransformer
)


import os
from incremental_module_loader import IncrementalModuleLoader

from .auth import AuthModule
from .config import FileConfigStorageProvider, ConfigModule
from .logs import _logger, setup_logging
from .storage import SITE_DATA_DIR
from .spinhub import SpinHubClientModule

from .services.auth0 import SpinHubAuthProvider

logger = _logger('root')

def Spintop(verbose=False, org_id=None, spinhub_url=None):
    # Default factory args
    return SpintopFactory(ephemeral_config=dict(
        verbose=verbose,
        org_id=org_id,
        spinhub_url=spinhub_url
    ))

def SpintopFactory(
            ephemeral_config={},
            logs_factory=setup_logging,
            config_factory=ConfigModule,
            config_storage_factory=FileConfigStorageProvider,
            auth_provider_factory=SpinHubAuthProvider,
            auth_factory=AuthModule, 
            spinhub_factory=SpinHubClientModule,
            final_factory=None
        ):
    
    loader = IncrementalModuleLoader()
    loader.update(
        ephemeral_config=ephemeral_config,
        config_storage=FileConfigStorageProvider(os.path.join(SITE_DATA_DIR, 'spintop.yml'))
    )
    
    loader.load(logs=logs_factory)
    
    config = loader.load(config=config_factory)
    # config.load(ephemeral_config)
    
    loader.load(auth_provider=auth_provider_factory)
    loader.load(auth=auth_factory)
    loader.load(spinhub=spinhub_factory)
    spintop_or_final = loader.load(spintop=SpintopModule)
    
    if final_factory:
        spintop_or_final = loader.load(final_factory)
    
    return spintop_or_final

def SpintopWorkerFactory(worker_cls, **factory_kwargs):
    worker = SpintopFactory(final_factory=worker_cls, **factory_kwargs)
    return worker
    

class SpintopModule(object):
    def __init__(self, config, spinhub, auth):
        self.config = config
        self.spinhub = spinhub
        self.auth = auth
    
    def assert_no_login(self):
        self.auth.assert_no_login()
    
    def login(self, username, password):
        self.assert_no_login()
        self.auth.login(username, password)
        self.auth.save_credentials()
    
    def stored_logged_username(self):
        if self.auth.credentials:
            return self.auth.credentials.get('username')
        else:
            return None
    
    def logout(self):
        return self.auth.logout()
    
    def delete_config(self):
        self.logout()
        self.config.delete_config()
    
    def register_machine(self, name, token, org_id=None):
        return self.spinhub.register_machine(name, token, org_id=org_id)
    
    def get_user_orgs(self):
        user_info = self.spinhub.user_info
        return user_info['organizations']


        
        
    
    
        


from functools import wraps

from pymongo import UpdateOne, MongoClient, ASCENDING, DESCENDING
from pymongo.errors import BulkWriteError

from .queryset import MongoQuery

def db_from_mongo_uri(mongo_uri, database_name):
    client = MongoClient(mongo_uri)
    return client[database_name]


class MongoOperations(object):
    def __init__(self, mongo_collection):
        self.ops = mongo_collection
        
    def find(self, query):
        query_dict = MongoQuery(query).build()
        return self.ops.find(query_dict)
        
    def insert_many(self, objs):
        return self.ops.insert_many(objs)
    
    def update_many(self, objs, by='_id'):
        updates = [
            UpdateOne({by: obj[by]}, {'$set': obj}) for obj in objs
        ]
        return self.ops.bulk_write(updates)
    
    def delete_many(self, query):
        query_dict = MongoQuery(query).build()
        return self.ops.delete_many(query_dict)
        
    def create_index(self, fields_and_direction, unique=False):
        self.ops.create_index(fields_and_direction, unique=unique)


from spintop.generators import Generator

from spintop.models import (
    TestRecordSummary, 
    FeatureRecord, 
    SpintopTestRecord
)

from ..logs import _logger

logger = _logger('persistence')


class MissingMapper(Exception):
    def __init__(self, cls):
        super().__init__("Mapper for class {!r} is mandatory.".format(cls))

class NoMapperForObject(Exception):
    def __init__(self, obj, mappers):
        super(NoMapperForObject, self).__init__(
            'There are no known mapper able to interact with obj {!r} of class {}. Declared mappers are: {}'.format(
                obj,
                obj.__class__,
                [cls.__name__ for cls in mappers]
            )
        )
        
class DuplicateMapperClassName(Exception):
    def __init__(self, objcls):
        super(DuplicateMapperClassName, self).__init__(
            'The name of the class {} is duplicate. The class name linked to a mapper must be unique.'.format(
                objcls,
            )
        )

class PersistenceFacade(object):
    logger = logger
    def __init__(self, mappers):
        self.mappers = mappers
        self.mappers_name_index = create_mapper_name_index(mappers)
        self._init()
        
    def create(self, records):
        raise NotImplementedError()
        
    def retrieve(self, test_selector=None, feature_selector=None):
        """Generator."""
        raise NotImplementedError()
        
    def update(self, records):
        raise NotImplementedError()
    
    def delete(self, match_query):
        raise NotImplementedError()

    def _init(self):
        for key, mapper in self.mappers.items():
            mapper.init()
        
    def _get_mapper(self, obj):
        try:
            mapper = self.mappers[obj.__class__]
        except KeyError:
            raise NoMapperForObject(obj, mappers=self.mappers)
        return mapper
        
    def __getattr__(self, mapper_cls_name):
        try:
            return self.mappers_name_index[mapper_cls_name]
        except KeyError:
            raise AttributeError(mapper_cls_name)
    
    def create_records_generator(self):
        return PersistenceGenerator(self)
        
def create_mapper_name_index(mappers):
    mappers_name_index = {}
    for mapped_cls, mapper in mappers.items():
        name = mapped_cls.__name__
        if name in mappers_name_index:
            raise DuplicateMapperClassName(mapped_cls)
        mappers_name_index[name] = mapper
    return mappers_name_index
    
class PersistenceGenerator(Generator):
    def __init__(self, facade):
        super().__init__()
        self.facade = facade
    
    def __call__(self, test_selector=None, feature_selector=None):
        return self.facade.retrieve(
            test_selector=test_selector, 
            feature_selector=feature_selector
        )

class Mapper(object):
    def init(self):
        pass


import warnings
import types

class TransformError(Exception): pass

class Transformer(object):
    def __call__(self, *args, **kwargs):
        raise NotImplementedError()
    
    def generate(self, *args, **kwargs):
        for value in self(*args, **kwargs):
            yield value
    
    def append(self, *transforms):
        return join_transforms(self, *transforms)

    def __add__(self, transform):
        return self.join_with(transform)

    def __radd__(self, other):
        return other.join_with(self)
    
    def __rshift__(self, other):
        return self + other
        
    def __rrshift__(self, other):
        return other + self
    
    def join_with(self, other):
        return join_transforms(self, other)

class FunctionTransformer(Transformer):
    def __init__(self, fn):
        self.fn = fn
        
    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

class TransformBuilder():
    def __init__(self, *transforms, _first_step=None):
        self.first_step = _first_step
        self.append(*transforms)
            
    @property
    def last_step(self):
        last_step = self.first_step
        while last_step is not None and last_step.next_step:
            last_step = last_step.next_step
        return last_step
            
    def append(self, *transforms):
        if len(transforms) == 1 and isinstance(transforms[0], TransformStep):
            first_next_step = transforms[0]
        else:
            first_next_step = None
            for transform in reversed(transforms):
                first_next_step = TransformStep(transform, next_step=first_next_step)
        
        if self.first_step:
            self.last_step.set_next(first_next_step)
        else:
            self.first_step = first_next_step
        
    def prepend(self, transform):
        if self.first_step is not None:
            # Not first step
            self.first_step = self.first_step.prepend(transform)
        else:
            # First step
            self.first_step = TransformStep(transform)
            
    def __call__(self, *args, **kwargs):
        return self.first_step(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.first_step, name)
    
    def split(self, branches=2):
        return tuple(self.copy() for _ in range(branches))
        
    def copy(self):
        return TransformBuilder(_first_step=self.first_step.copy())
    
    def join_with(self, other):
        new_self = self.copy()
        new_self.append(other)
        return new_self

    def __repr__(self):
        return repr(self.first_step)

class TransformStep(Transformer):
    def __init__(self, transform_fn, next_step=None):
        if isinstance(transform_fn, TransformStep):
            raise TransformError("The transform function cannot be an instance of TransformStep.")
        self._transform_fn = transform_fn
        self._next_step = next_step
    
    @property
    def next_step(self):
        return self._next_step
    
    @property
    def transform_fn(self):
        return self._transform_fn
    
    def __call__(self, *args, **kwargs):
        """ The first transform step can be called with as many arguments as desired. However, subsequent steps
        will receive only one argument: the result of the previous transform.
        """
        generator = self.generate(*args, **kwargs)
        result = list(generator)

        if len(result) != 1:
            raise Exception(
                "Transform did not return exactly one (!={}) value. "
                "If this is a generator and you wish to iterate the "
                "yield values, use the 'generate' method instead of "
                "__call__.".format(len(result)))
        else:
            return result[0]

    def generate(self, *args, **kwargs):
        obj = self.transform_fn(*args, **kwargs)
        
        if obj is None:
            raise TransformError("Transformer fn {} returned None".format(self.transform_fn))

        if isinstance(obj, types.GeneratorType):
            for subobj in obj:
                yield from self._call_next_step(subobj)
        else:
            yield from self._call_next_step(obj)

    def _call_next_step(self, obj):
        if self.next_step:
            yield from self.next_step.generate(obj)
        else:
            yield obj
        
    def set_next(self, next_step):
        if self.next_step:
            raise TransformError('Next step already defined ! This would break the transform pipeline.')
        self._next_step = next_step
        
    def prepend(self, transform):
        if isinstance(transform, TransformStep):
            previous_step = transform.copy(next_step=self)
        else:
            previous_step = self.__class__(transform, next_step=self)
        return previous_step
    
    def copy(self, next_step=None):
        next_step = next_step
        if next_step is None and self.next_step:
            next_step = self.next_step.copy()
        return self.__class__(self.transform_fn, next_step=next_step)
    
    def __repr__(self):
        return "TransformStep(fn={}), next=\n{}".format(self.transform_fn, repr(self.next_step))
    
class EmptyTransformer(Transformer):
    def __call__(self, arg):
        return arg
    
def join_transforms(*transforms):
    return TransformBuilder(*transforms)

def transformer(fn):
    return FunctionTransformer(fn)


import click
import requests

from pprint import pprint

from ..errors import SpintopException, AuthUnauthorized
from .schemas import credentials_schema

class AuthException(SpintopException):
    pass

REFRESH_MODULE = 'user'

class AuthModule(object):
    def __init__(self, auth_provider=None, config=None):
        self.auth_provider = auth_provider
        self.config = config
        
        self._credentials = None
        self._initialized = False
    
    @property
    def credentials(self):
        self._init_check()
        return self._credentials
    
    @credentials.setter
    def credentials(self, value):
        self._credentials = value
    
    def _init_check(self):
        if not self._initialized:
            self.credentials = self.config.get_credentials()
            # Other init stuff should go here
            self._initialized = True
    
    def assert_credentials(self):
        if not self.credentials:
            raise AuthException("No credentials available. Please login.")
    
    def assert_no_login(self):
        if self.credentials:
            raise AuthException("""
You already have credentials stored for %s. \
Spintop currently only support a single user per PC. \
Please logout before logging back in.
""" % self.credentials.username)
    
    def login(self, username, password):
        self.assert_no_login()
        
    def authenticate(self, username, password, scopes=[]):
        content = self.auth_provider.authenticate(username, password, scopes=['openid', 'email', 'profile', 'authorization'])
        self.credentials = credentials_schema.load({
            'username': username,
            'access_token': content.get('access_token'),
            'refresh_token': content.get('refresh_token'),
            'org_id': None,
            'refresh_module': REFRESH_MODULE
        })
        self.save_credentials()
        
    @property
    def credentials_key(self):
        
        key = self.credentials.get('username', None)
            
        if key is None:
            key = self.credentials.get('org_id', None)
            
        return key
        
    def save_credentials(self):
        self.config.set_credentials(self.credentials_key, self.credentials)
        self.config.save()
    
    def refresh_credentials(self):
        try:
            self.credentials = self._refresh_credentials_obj(self.credentials)
        except AuthUnauthorized:
            raise AuthException('Unable to refresh credentials.')
        self.save_credentials()
        
    def logout(self):
        # Add SpinHub-related logout stuff
        if self.credentials:
            self.auth_provider.revoke_refresh_token(self.credentials['refresh_token'])
            
        self.config.remove_credentials(self.credentials_key)
        self.config.save()
    
    def _refresh_credentials_obj(self, credentials):
        # username and refresh_token both stay valid
        if credentials is None:
            raise AuthUnauthorized()
            
        username = credentials['username']
        refresh_token = credentials['refresh_token']
        
        self._assert_refresh_token_exists(credentials)
        content = self.auth_provider.refresh_access_token(refresh_token)
        
        return credentials_schema.load({
            'username': username,
            'refresh_token': refresh_token,
            'access_token': content.get('access_token'),
            'org_id': None,
            'refresh_module': REFRESH_MODULE
        })

    def _assert_refresh_token_exists(self, credentials):
        if not credentials or not credentials['refresh_token']:
            raise AuthUnauthorized('No refresh token exists; unable to execute operation')
        
    def get_auth_headers(self):
        if self.credentials:
            return {
                'Authorization': 'Bearer ' + self.credentials.get('access_token')
            }
        else:
            return {}
        
    




from .operations import MongoOperations, ASCENDING
from .serialization import DataClassSerializer

from spintop.models import (
    TestRecordSummary, 
    FeatureRecord, 
    SpintopTestRecord, 
    SpintopFlatTestRecordBuilder,
    Query
)


from ...logs import _logger


SIMPLE_MAPPER_DATACLASSES_COLLECTION_NAMES = {}

SIMPLE_MAPPER_DATACLASSES_INIT_FNS = {}

logger = _logger('mongo')

def register_simple_mapper(cls, collection_name):
    SIMPLE_MAPPER_DATACLASSES_COLLECTION_NAMES[cls] = collection_name
    SIMPLE_MAPPER_DATACLASSES_INIT_FNS[cls] = None
    def _register_init(fn):
        SIMPLE_MAPPER_DATACLASSES_INIT_FNS[cls] = fn
        return fn
    return _register_init

# Mappers
@register_simple_mapper(TestRecordSummary, 'test-records')
def init_test_record_mapper(mapper):
    mapper.ops.create_index([("test_id.test_uuid", ASCENDING)], unique=True)
    
register_simple_mapper(FeatureRecord, 'test-features')

def create_mappers(mongo_db, serializer=None, operations_factory=None):
    if operations_factory is None: operations_factory = MongoOperations
    
    # Simple mappers are 1:1 with a document in a mongo collection.
    mappers = {}
    for dataclass, collection_name in SIMPLE_MAPPER_DATACLASSES_COLLECTION_NAMES.items():
        init_fn = SIMPLE_MAPPER_DATACLASSES_INIT_FNS[dataclass]
        ops = operations_factory(mongo_db[collection_name])
        mappers[dataclass] = SimpleMongoMapper(ops, serializer=serializer, on_init=init_fn, name=collection_name)
    
    return mappers

class SimpleMongoMapper(object):
    def __init__(self, ops, serializer=None, on_init=None, name="simple"):
        if serializer is None: serializer = DataClassSerializer()
        self.ops = ops
        self.dc_serializer = serializer
        self.on_init = on_init
        self.logger = logger.getChild(name)
    
    def init(self):
        if self.on_init: self.on_init(self)
    
    def create(self, dataclasses):
        serialized_dcs = self.dc_serializer.serialize_many(dataclasses, keep_id_if_none=False)
        
        resp = self.ops.insert_many(serialized_dcs)
        self.logger.info('Inserted %d new documents' % len(resp.inserted_ids))
        
        # ids are inserted client side into the 'serialized' dictionnary.
        # We can re-insert _ids into dataclasses.
        for dc, serialized in zip(dataclasses, serialized_dcs):
            dc.oid = serialized['_id']
            
    def retrieve(self, query):
        responses = self.ops.find(query)
        return self.dc_serializer.deserialize_many(responses)
        
    def update(self, dataclasses):
        serialized = self.dc_serializer.serialize_many(dataclasses)
        self.ops.update_many(serialized, by='_id')
        
    def delete(self, query):
        resp = self.ops.delete_many(query)
        self.logger.info('Deleted %d documents' % resp.deleted_count)


import os
import appdirs

from tempfile import gettempdir

APPNAME = 'spintop'
APPAUTHOR = 'tackv'

SITE_DATA_DIR = appdirs.user_data_dir(appname=APPNAME, appauthor=APPAUTHOR)

if not os.path.exists(SITE_DATA_DIR):
    os.makedirs(SITE_DATA_DIR)  
    
TEMP_DIR = os.path.join(gettempdir(), 'spintop')



from copy import deepcopy
from ..errors import SpintopException

from ..logs import _logger

logger = _logger('config')

LOADED_CONFIG = None

class ConfigModule(object):
    def __init__(self, config_storage, ephemeral_config):
        
        self.config_storage = config_storage
        self.ephemeral_config = ephemeral_config
        
        self._config = {}
        self._config_loaded = False
        
        self._profile = None
    
    @property
    def config(self):
        if not self._config:
            self.load(self.ephemeral_config)
        return self._config
    
    def delete_config(self):
        self._config = self.create_empty()
        self.config_storage.delete_spintop_config()
        
    def create_empty(self, default_url=None, profile_name=None, **other_profile_config):
        
        if default_url is None:
            default_url = 'https://dev-api.spinhub.ca'
            
        if profile_name is None:
            profile_name = 'default'
        
        self._config = dict(
            default_profile = profile_name,
            profiles = {},
            credentials = {}
        )
        
        self.create_profile(default_url=default_url, profile_name=profile_name, **other_profile_config)
        
        return self._config
    
    def create_profile(self, machine_id=None, default_url=None, spinhub_url=None, profile_name=None, org_id=None, **_):
        if not spinhub_url:
            spinhub_url = default_url
        
        profile = dict(
            credentials = None,
            spinhub_url = spinhub_url
        )
        self._config['profiles'] = self._config.get('profiles', {})
        self._config['profiles'][profile_name] = profile
        logger.info('Created a profile named %s' % profile_name)
        return profile
    
    
    def update_profile(self, profile_name=None, **kwargs):
        if profile_name is None:
            profile = self._profile
        else:
            profile = self.get_profile(profile_name)
            
        profile.update(kwargs)
    
    def set_credentials(self, key, credentials):
        self._config['credentials'] = self._config.get('credentials', {})
        self._config['credentials'][key] = credentials
        
        profile = self.get_selected_profile()
        if profile.get('credentials_key') is None:
            profile['credentials_key'] = key
            
        
    def remove_credentials(self, key):
        if key in self._config.get('credentials', {}):
            del self._config['credentials'][key]
    
    def get_credentials(self, key=None):
        if key is None:
            key = self.get_selected_profile().get('credentials_key')
        
        if key is None:
            return None
        
        return self._config.get('credentials', {}).get(key, {})
    
    def get_selected_profile(self):
        return self.get_profile()
        
    def load(self, ephemeral_config):
        config = self.get_stored()
        before_load = deepcopy(config)
        
        if not config:
            config = self.create_empty(**ephemeral_config)
        
        self._profile = self.get_profile(**ephemeral_config)
        
        # logger.info('Loaded profile named %s' % self._profile['name'])
        
        if config != before_load:
            self.save()
        
    def get_profile(self, profile_name=None, **other_profile_config):
        if profile_name is None:
            profile_name = self.config.get('default_profile')
        
        profile = self._config.get('profiles', {}).get(profile_name, None)
        
        if profile is not None:
            return profile
        
        # profile not found
        profile = self.create_profile(profile_name=profile_name, **other_profile_config)
        self.save()
        return profile
            
    def get_stored(self):
        if not self._config_loaded:
            self._config = self.config_storage.retrieve_spintop_config()
            self._config_loaded = True
        return self._config
    
    def save(self):
        self._config = self.config_storage.store_spintop_config(self._config)
        self._config_loaded = True
        return self._config
    


""" Storage-friendly internal data classes. """
from collections import Mapping
from base64 import urlsafe_b64decode, urlsafe_b64encode
from uuid import UUID, uuid4

from datetime import datetime
from dataclasses import dataclass, fields, _MISSING_TYPE, MISSING, field, asdict
from typing import Union, List

from .meta import (
    BaseDataClass, 
    register_type, 
    KeyValueOf,
    _fields_cache,
    model_dataclass
)

from .view import ComplexPrimitiveView

NAME_SEPARATOR = ':'
NO_VERSION = '0'

def create_uuid():
    return uuid4()

def uuid_to_slug(uuidstring):
    try:
        uuid = UUID(uuidstring)
    except AttributeError:
        uuid = uuidstring
    
    return urlsafe_b64encode(uuid.bytes).rstrip(b'=').decode('ascii')

def slug_to_uuid(slug):
    return str(UUID(bytes=urlsafe_b64decode(slug + '==')))

def is_type_of(cls, base_cls):
    return cls._type == base_cls._type

@model_dataclass()
@register_type('dut')
class DutIDRecord(BaseDataClass):
    id: str
    version: str
    
    @classmethod
    def create(cls, id_or_dict, version=None):
        if isinstance(id_or_dict, Mapping):
            id = id_or_dict.get('id', None)
            version = id_or_dict.get('version', None)
        else:
            id = id_or_dict
            
        if version is None:
            version = NO_VERSION
            
        return cls(
            id=id,
            version=version
        )
        
    def match(self, other):
        id_match = self.id == other.id if self.id is not None else True
        version_match = self.version == other.version if self.version is not None else True
        return id_match and version_match

@model_dataclass()
@register_type('test_id')
class TestIDRecord(BaseDataClass):
    testbench_name: str
    dut: DutIDRecord
    test_uuid: str
    tags: dict
    start_datetime: datetime
    
    @classmethod
    def create(cls, testbench_name, dut, start_datetime, tags=None):
        return cls(
            testbench_name=testbench_name,
            dut=DutIDRecord.create(dut),
            test_uuid=uuid_to_slug(create_uuid()),
            start_datetime=start_datetime,
            tags= tags if tags else {}
        )
        
    
    def add_tag(self, key, value=True):
        self.tags[key] = value
        
    def remove_tag(self, key):
        del self.tags[key]
        
@model_dataclass()
@register_type('features')
class FeatureRecordNoData(BaseDataClass):
    oid: str
    test_id: TestIDRecord
    name: str
    version: int
    depth: int
    index: int
    ancestors: List[str]
    original: bool
    
    @classmethod
    def get_data_fields(cls):
        """ The data fields are all fields except the ones declare up to this cls
        in the cls MRO.
        
        If a subclass defines only a dataclass field 'foo', this method will return
        foo only as a data field.
        """
        return set(_fields_cache.retrieve(cls)) - set(_fields_cache.retrieve(FeatureRecordNoData))
    
    @property
    def complete_name(self):
        return NAME_SEPARATOR.join(self.ancestors + [self.name])
        
    def as_dict(self):
        """ Transforms this object recursively into a dict. Adds the static
        _type field to the data.."""
        data = super(FeatureRecordNoData, self).as_dict()
        return data
    
@model_dataclass()
class FeatureRecord(FeatureRecordNoData):
    user_data: dict # TODO Replace with data_view with descriptor protocol
    
    @classmethod
    def defaults(cls, **others):
        return super(FeatureRecord, cls).defaults(
            user_data={}, 
            **others
        )
    
@model_dataclass()
@register_type('outcome')
class OutcomeData(BaseDataClass):
    is_pass: bool
    is_skip: bool = False
    is_abort: bool = False
    
    def __bool__(self):
        return self.is_pass and not self.is_abort



@model_dataclass()
@register_type('features_o')
class FeatureOutcomeRecord(FeatureRecord):
    outcome: OutcomeData

@model_dataclass()
@register_type('phases')
class PhaseFeatureRecord(FeatureOutcomeRecord):
    duration: float
    
@model_dataclass()
@register_type('measures')
class MeasureFeatureRecord(FeatureOutcomeRecord):
    value: Union[float, str]
    
@model_dataclass()
@register_type('template')
class FeatureRecordTemplate(FeatureRecord):
    source: FeatureRecord
    records_count: int = 0
    
    def increment_records_count(self):
        self.records_count = self.records_count + 1
    
    
@model_dataclass()
@register_type('top_level')
class TestRecordSummary(PhaseFeatureRecord):
    total_feature_count: int = 0
    template_match_score: float = None
    template_match_score_normalized: float = None
    
    @classmethod
    def get_data_fields(cls):
        base = super(TestRecordSummary, cls).get_data_fields()
        return list(base) + [FeatureRecord.test_id]
    
@model_dataclass()
class DutOp(BaseDataClass):
    dut_match: DutIDRecord
    dut_after: DutIDRecord
    op_datetime: datetime
    
    @classmethod
    def create(cls, dut_match, dut_after, op_datetime):
        return cls(
            dut_match=DutIDRecord.create(dut_match),
            dut_after=DutIDRecord.create(dut_after),
            op_datetime=op_datetime
        )
    
    def does_applies_to(self, dut, on_datetime):
        return self.dut_match.match(dut) and self.op_datetime < on_datetime
    
    def apply_or_return(self, dut, on_datetime):
        if self.does_applies_to(dut, on_datetime):
            return self.apply(dut)
        else:
            return dut
        
    def apply(self, dut):
        """Apply op to dut. Does not check for datetime. """
        new_dut = self.dut_after.copy()
        if new_dut.id is None:
            new_dut.id = dut.id
        return new_dut
        
        
def default_cls_missing_view_fn(cls):
    return cls.get_default_view()
        
class DefaultPrimitiveView(ComplexPrimitiveView):
    def __init__(self):
        super(DefaultPrimitiveView, self).__init__(
            BaseDataClass, 
            cls_missing_view_fn=default_cls_missing_view_fn
        )
        
        # summary_mapping = TestRecordSummary.get_default_view_mapping()
        
        # self.add_view(
        #     TestRecordSummary,
        #     summary_mapping
        # )


import os
import sys
import click
import traceback
import logging
import threading
import time
from pprint import pformat, pprint
from tabulate import tabulate

from . import logs, storage
from .auth import AuthModule
from .compat import format_exception
from .errors import SpintopException
from .models import Query
from .spintop import Spintop


class CLIParseException(SpintopException):
    pass

logger = logs._logger('cli')

_spintop = None 
spintop_kwargs = {}
def spintop():
    global _spintop, spintop_kwargs
    if _spintop is None:
        _spintop = Spintop(**spintop_kwargs)
    return _spintop

@click.group()
@click.option('-v', '--verbose', is_flag=True, default=False)
@click.option('--url', default='https://api.spinhub.ca')
@click.option('--org-id', default=None)
def _cli(verbose, url, org_id):
    global spintop_kwargs
    logs.VERBOSE = verbose
    spintop_kwargs = dict(verbose=verbose, org_id=org_id, spinhub_url=url)

@_cli.command('site-data-dir')
def site_data_dir():
    click.echo(storage.SITE_DATA_DIR)
    
@_cli.command('login')
@click.option('--username', default=None)
@click.option('--password', default=None)
def login_command(username, password):
    if username is None and password is not None:
        raise CLIParseException('You cannot specify the password as option but not the username')
    
    spintop().assert_no_login() # Check first before prompting
    
    if username is None:
        username = click.prompt('Username')
        
    if password is None:
        password = click.prompt('Password', hide_input=True)
    
    spintop().login(username, password)

@_cli.command('register')
@click.argument('name')
@click.option('--token', default=None)
@click.option('--org-id', default=None)
def register(name, token, org_id=None):
    spintop().register_machine(name, token, org_id=org_id)

@_cli.command('logout')
def logout_command():
    stored_username = spintop().stored_logged_username()
    if stored_username:
        if click.confirm('Logging out %s. Continue ?' % stored_username):
            spintop().logout()
    else:
        click.echo('No credentials stored.')
    
@_cli.command('show-token')
def get_org_token():
    print(spintop().spinhub.get_token())

@_cli.command('ls-orgs')
def get_orgs():
    orgs = spintop().get_user_orgs()
    spintop().spinhub.retrieve_tests(test_selector=Query().testbench_name_is('demo-testbench'))
    data = [[org['uuid'], org['key']] for org in orgs]
    click.echo(tabulate(data, headers=['UUID', 'Org Key']))

@_cli.command('ls-peers')
def get_orgs():
    this_machine = spintop().get_machine()
    peers = spintop().get_peers()
    data = [[
            peer_machine['uuid'], 
            peer_machine['nickname'] + (' (self)' if peer_machine['uuid'] == this_machine['uuid'] else ''),
            'Machine' if peer_machine['is_machine'] else 'User'
        ] for peer_machine in peers]
    click.echo(tabulate(data, headers=['UUID', 'Nickname', 'Type']))
    
@_cli.group('config')
def config():
    pass
    
@config.command('show')
def show_config():
    config = spintop().config.get_stored()
    click.echo(pformat(config))
    
@config.command('update')
@click.argument('key_values', nargs=-1)
def update_profile(key_values):
    profile = spintop().config.get_selected_profile()
    updates = {}
    for key_value in key_values:
        key, value = key_value.split('=')
        updates[key] = value
        
    click.echo('Will update profile named %s with:' % profile['name'])
    click.echo(pformat(updates))
    if click.confirm('Proceed ?'):
        spintop().config.update_profile(**updates)
        spintop().config.save()
    
@config.command('clear')
def clear_config():
    if click.confirm('This will delete all local credentials and profiles. Proceed ?'):
        spintop().delete_config()
    
@_cli.command('test-api-private')
def test_private_command():
    print(spintop().spinhub.test_private_endpoint().content)
    
@_cli.command('force-corrupt-access-token')
def force_corrupt_access_token():
    spintop().spinhub.auth.credentials.access_token = spinhub.auth.credentials.access_token[0:-4]
    spintop().spinhub.auth.save_credentials()


def cli(args=None, internal=True):
    try:
        return _cli(args, prog_name='spintop')
    except Exception as e:
        if logs.VERBOSE:
            click.echo(traceback.format_exc())
        else:
            click.echo(format_exception(e))
    except SystemExit:
        if not internal:
            raise
            
if __name__ == '__main__':
    cli(internal=False)


from datetime import datetime
from time import mktime
from copy import copy

from dataclasses_serialization.serializer_base import dict_to_dataclass, dict_serialization, list_deserialization
from dataclasses_serialization.bson import BSONSerializer
from dataclasses_serialization.json import JSONSerializer

from .internal import BaseDataClass
from .collection import SpintopTestRecord

SpintopBSONSerializer = None
SpintopJSONSerializer = None

def get_serializer():
    return get_bson_serializer()

def get_bson_serializer():
    """Make or return the singleton SpintopBSONSerializer"""
    global SpintopBSONSerializer
    if SpintopBSONSerializer:
        return SpintopBSONSerializer

    serializer = copy(BSONSerializer)
    _make_serializer(serializer)
    SpintopBSONSerializer = serializer
    return serializer

def get_json_serializer():
    global SpintopJSONSerializer
    if SpintopJSONSerializer:
        return SpintopJSONSerializer
    
    serializer = copy(JSONSerializer)
    
    @serializer.register_serializer(datetime)
    def serialize_datetime(obj):
        return mktime(obj.timetuple())

    @serializer.register_deserializer(datetime)
    def deserialize_datetime(cls, obj):
        return datetime.fromtimestamp(obj)
        
    _make_serializer(serializer)

    SpintopJSONSerializer = serializer
    return serializer

def _make_serializer(serializer):
    orig_deserialize = serializer.deserialize
    def wrapped_deserialize(cls, obj):
        """ Always support any obj that is None. Default implementation
        will raise an error if obj is None and the type of a dataclass
        field is anything but Any or None
        """
        if obj is None:
            return None
        else:
            return orig_deserialize(cls, obj)
    
    serializer.deserialize = wrapped_deserialize
    
    @serializer.register_serializer(BaseDataClass)
    def serialize(obj):
        """Custom serializer for BaseDataClass that extracts the static
        attribute '_type' into the serialized dict for future deserialization."""
        data = dict_serialization(dict(obj.__dict__), key_serialization_func=serializer.serialize, value_serialization_func=serializer.serialize)
        data.update({'_type': obj._type})
        return data

    @serializer.register_deserializer(BaseDataClass)
    def deserialize(cls, obj):
        """ Custom deserializer that infers the BaseDataClass subclass based
        on the _type attribute. 
        """
        # Strips the _type field and returns the sub cls, if found. Else returns the cls
        # as passed in parameter
        cls, obj = BaseDataClass.cls_data_from_dict(cls, obj)
        return dict_to_dataclass(cls, obj, deserialization_func=serializer.deserialize)

    
    @serializer.register_serializer(SpintopTestRecord)
    def serialize_spintop_test_record(obj):
        return serializer.serialize(obj.as_dict())

    @serializer.register_deserializer(SpintopTestRecord)
    def deserialize_spintop_test_record(cls, obj):

        if 'test_record' in obj:
            obj['test_record'] = serializer.deserialize(BaseDataClass, obj['test_record'])
        
        if 'features' in obj:
            obj['features'] = [serializer.deserialize(BaseDataClass, feat) for feat in obj['features']]

        return cls.from_dict(obj)

    
    return serializer



from collections import Mapping, OrderedDict

class _KeyValueOfAccessor(object):
    def __init__(self, attr_name):
        self.attr_name = attr_name
    
    def __call__(self, obj):
        return getattr(obj, self.attr_name)
    
class _ValueAccessor(object):
    def __init__(self, value):
        self.value = value
    
    def __call__(self, obj):
        return self.value
    
    
class DataClassPrimitiveView(object):
    def __init__(self, key_values_mapping={}):
        self.build_accessors(key_values_mapping)
        
    def build_accessors(self, key_values_mapping):
        """ Creates a dict of callable
        """
        self._accessors_mapping = {_build_single_accessor(key): _build_single_accessor(value) for key, value in key_values_mapping.items()}
        
    def apply(self, data_obj, flatten_dict=False):
        result = {}
        for accessor_key, accessor_value in self._accessors_mapping.items():
            key = _apply_single_accessor(accessor_key, data_obj)
            value = _apply_single_accessor(accessor_value, data_obj)
            
            if isinstance(key, tuple) and not flatten_dict:
                # Push value deep into a nested dictionnary
                # key = (x, y, z)
                # result.update({x: {y: {z: value}}})
                result.update(deepen_value_in_dict(key, value))
            else:
                # or keep as is (flat)
                # key = (x, y, z)
                # result[(x, y, z)] = value
                result[key] = value
            
        return result

def _build_single_accessor(obj):
    if isinstance(obj, tuple):
        return tuple(_build_single_accessor(subobj) for subobj in obj)
    elif hasattr(obj, 'name'):
        return _KeyValueOfAccessor(obj.name) # Will return the value of the attribute named obj.name
    elif callable(obj):
        return obj # Obj is an accessor
    else:
        return _ValueAccessor(obj) # Will always return obj
    
def _apply_single_accessor(accessor, obj):
    if isinstance(accessor, tuple):
        return tuple(_apply_single_accessor(subaccessor, obj) for subaccessor in accessor)
    else:
        return accessor(obj)
        
    
def default_cls_missing_view_fn(cls):
    raise NotImplementedError()

class ComplexPrimitiveView(object):
    def __init__(self, base_cls, cls_missing_view_fn=None):
        if cls_missing_view_fn is None: cls_missing_view_fn = default_cls_missing_view_fn
        
        self.base_cls = base_cls
        self.cls_missing_view_fn = cls_missing_view_fn
        self.custom_views = {}
        
    def _find_closest_parent(self, cls):
        closest_parent = None
        
        if cls in self.custom_views:
            return cls
        
        for match_cls in self.custom_views:
            possible_parent = None
            if issubclass(cls, match_cls):
                possible_parent = match_cls
            
            if possible_parent and (closest_parent is None or issubclass(possible_parent, closest_parent)):
                # possible parent is closer to cls than the previous closest parent.
                closest_parent = possible_parent
                
        return closest_parent
    
    def _find_view(self, cls):
        best_match = self._find_closest_parent(cls)
        if best_match:
            return self.custom_views[best_match]
        else:
            return self.cls_missing_view_fn(cls)
    
    def add_view(self, cls, view_mapping):
        self.custom_views[cls] = DataClassPrimitiveView(view_mapping)
        
    def apply(self, root_data_obj, flatten_dict=False, key_prefix=None):
        flatten_key_prefix = None
        if flatten_dict:
            if key_prefix is None:
                flatten_key_prefix = ()
            else:
                flatten_key_prefix = key_prefix
                
        result = self._apply_data_obj(root_data_obj, flatten_key_prefix=flatten_key_prefix)
        
        if not flatten_dict and key_prefix:
            return deepen_value_in_dict(key_prefix, result)
        else:
            return result
    
        
    
    def _apply_data_obj(self, data_obj, flatten_key_prefix=None):
        mapping = self._find_view(data_obj.__class__).apply(data_obj, flatten_dict=flatten_key_prefix is not None)
        
        flatten_key_prefix_built = _build_single_accessor(flatten_key_prefix)
        flatten_key_prefix_applied = _apply_single_accessor(flatten_key_prefix_built, data_obj)
        
        return self._apply_mapping(mapping, flatten_key_prefix=flatten_key_prefix_applied)
    
    def _apply_mapping(self, mapping, flatten_key_prefix=None):
        result = OrderedDict()
        
        for key, value in mapping.items():
            
            flat_key = None
            if flatten_key_prefix is not None:
                flat_key = join_key_prefix(flatten_key_prefix, key)
            
            discard_false_subvalue = False 
            
            if isinstance(value, self.base_cls):
                subvalue = self._apply_data_obj(value, flatten_key_prefix=flat_key)
                # If bool(subvalue) is False, then the sub-view returned an empty mapping or None. Do not consider it.
                discard_false_subvalue = True 
            elif isinstance(value, Mapping):
                subvalue = self._apply_mapping(value, flatten_key_prefix=flat_key)
                # If bool(subvalue) is False, the mapping is empty or none. Do not consider it.
                discard_false_subvalue = True 
            elif flat_key is not None:
                # limit case.
                # If flatten, we will .update the result instead of applying
                subvalue = {flat_key: value}
            else:
                subvalue = value
            
            if discard_false_subvalue and not subvalue:
                continue # discard
            
            if flat_key:
                for subkey in subvalue:
                    if subkey in result:
                        print("{} already in result.".format(subkey))
                result.update(subvalue)
            else:
                if key in result:
                    print("{} already in result.".format(key))
                result[key] = subvalue
        
        return result

def join_key_prefix(key_prefix, key):
    if isinstance(key, tuple):
        key = key_prefix + key
    else:
        key = key_prefix + (key,)
    return key

def deepen_value_in_dict(key_prefix, value):
    result = {}
    next_dict = result
    for key in key_prefix[:-1]:
        next_dict[key] = {}
        next_dict = next_dict[key]
    next_dict[key_prefix[-1]] = value
    
    return result


import os

from spintop.persistence import PersistenceFacade
from spintop.models import SpintopFlatTestRecordBuilder, Query

from .operations import db_from_mongo_uri
from .mappers import create_mappers, TestRecordSummary, FeatureRecord

class MongoPersistenceFacade(PersistenceFacade):
    def __init__(self, mongo_db, serializer=None):
        mappers = create_mappers(mongo_db, serializer=serializer)
        super(MongoPersistenceFacade, self).__init__(mappers=mappers)

        self.tr_mapper = mappers[TestRecordSummary]
        self.feature_mapper = mappers[FeatureRecord]
    
    @classmethod
    def from_mongo_uri(cls, mongo_uri, database_name):
        return cls(db_from_mongo_uri(mongo_uri, database_name))
    
    @classmethod
    def from_env(cls, mongo_uri_env_name='SPINTOP_MONGO_URI', database_name_env_name='SPINTOP_DATABASE_NAME'):
        return cls.from_mongo_uri(
            mongo_uri=os.environ[mongo_uri_env_name],
            database_name=os.environ[database_name_env_name]
        )
        
    def create(self, records):
        self._iter_op('create', records)
        
    def _iter_op(self, op_name, records):

        test_top_level_records = [record.data for record in records]
        all_features = []
        for record in records:
            all_features += record.features

        self.logger.info('Processing {} of {} flat records.'.format(
                op_name, len(test_top_level_records)
            )
        )
        getattr(self.tr_mapper, op_name)(test_top_level_records)
        getattr(self.feature_mapper, op_name)(all_features)
        
        
    def retrieve(self, test_selector=None, feature_selector=None):
        top_level_records = self.tr_mapper.retrieve(test_selector)
        
        test_uuid_to_flat_builder = {tlr.test_id.test_uuid: SpintopFlatTestRecordBuilder(data=tlr, features=[]) for tlr in top_level_records}
        
        selected_ids = [tlr.test_id.test_uuid for tlr in top_level_records]
        
        if not feature_selector: feature_selector = Query()
        
        feature_selector.test_uuid_any_of(selected_ids)
        
        all_features = self.feature_mapper.retrieve(feature_selector)
        
        for feature in all_features:
            test_uuid_to_flat_builder[feature.test_id.test_uuid].features.append(feature)
        
        for _, flat_builder in test_uuid_to_flat_builder.items():
            yield flat_builder.build()
        
    def update(self, records):
        self._iter_op('update', records)
    
    def delete(self, match_query):
        self.logger.info('Deleting flat records based on query: %s.' % repr(match_query))
        self.tr_mapper.delete(match_query)
        self.feature_mapper.delete(match_query)


from itertools import chain

from collections import Mapping, OrderedDict, defaultdict
from typing import NamedTuple, List

from .internal import TestIDRecord, TestRecordSummary, FeatureRecord, DefaultPrimitiveView
from .view import ComplexPrimitiveView

def normalize_features_len(dict_array):
    new_dict = {}
    max_column_key_len = max([len(key) for key in dict_array])
    for key, value in dict_array.items():
        if len(key) < max_column_key_len:
            key = key + ('',)*(max_column_key_len - len(key))
        new_dict[key] = value
    return new_dict
    
            
class SpintopTestRecordCollection(object):
    def __init__(self, flat_records=None):
        self.records = list(flat_records) if flat_records else []
    
    @property
    def collector(self):
        return self.add_record
    
    def add_record(self, record):
        self.records.append(record)
        
    def apply(self, *fns):
        for record in self.iter_records():
            for fn in fns:
                fn(record)
                
    def iter_records(self):
        for record in self.records:
            yield record
                
    def count_unique(self, key=lambda record: record.test_id.test_uuid):
        occurances = set()
        self.apply(lambda record: occurances.add(key(record)))
        return len(occurances)
            
    def sort(self, key=lambda record: record.test_id.start_datetime):
        self.records.sort(key=key)
        
    def avg_feature_count(self):
        feature_counts = [len(record.features) for record in self.iter_records()]
        return sum(feature_counts)/len(feature_counts)
    
class SpintopTestRecord(object):
    def __init__(self, test_record=None, features=None):
        self._data = None
        self._features = None
        
        self.data = test_record if test_record else TestRecordSummary.null() # The top-level data
        self.features = tuple(features) if features else tuple()
    
    @property
    def features(self):
        return self._features if self._features else tuple()
    
    @features.setter
    def features(self, features):
        self._broadcast_test_id(features)
        self._features = sorted(features, key=lambda feature: feature.index)
        
    def compute_stats(self):
        if not self._data.total_feature_count:
            self._data.total_feature_count = len(self.features)
        
    @property
    def data(self):
        return self._data
    
    @data.setter
    def data(self, value):
        self._data = value
        self._broadcast_test_id(self.features)
        
    @property
    def test_id(self):
        return self.data.test_id
    
    def __hash__(self):
        return hash(self.test_id.test_uuid)

    def __eq__(self, other):
        return self.test_id.test_uuid == other.test_id.test_uuid

    
    def remove_duplicate_features(self):
        feature_names = set()
        features = []
        for feature in self.features:
            if feature.name in feature_names:
                continue
            else:
                features.append(feature)
                feature_names.add(feature.name)
        
        self.features = features
    
    def _broadcast_test_id(self, features):
        """Uses the top level test_id and sets it for all features, ensuring
        that all features have the same TestIDRecord reference."""
        for feature in features:
            feature.test_id = self.data.test_id
    
    def reindex(self):
        for index, feature in enumerate(self.features):
            feature.index = index
            
    def fill_missing_from_source(self, fill_source, on_fill=None):
        max_current_feature_index = max(self.features, key=lambda f: f.index)
        new_features_len = len(fill_source)
        
        assert new_features_len >= max_current_feature_index.index, "Fill source must be same length or bigger than this test_record max index."
        
        new_features = [None]*new_features_len
        
        if on_fill is None:
            on_fill = lambda obj: obj.copy()
            
        def fill_range(start, end):
            for i in range(start, end):
                new_features[i] = on_fill(fill_source[i])
        
        current_index = 0
        for feature in sorted(self.features, key=lambda f: f.index):
            fill_range(current_index, feature.index)
            new_features[feature.index] = feature # Keep this feature
            current_index = feature.index + 1
        
        fill_range(current_index, new_features_len) # Fill end
        
        self.features = new_features
                
    def find_feature(self, condition, start_index=0):
        return next(feat for feat in self.features[start_index:] if condition(feat))
    
    def add_tag(self, key, value=True):
        self.test_id.add_tag(key, value=value)
        
    def remove_tag(self, key):
        self.test_id.remove_tag(key)

    def as_dict(self):
        return dict(
            test_record=self._data,
            features=self._features
        )

    @classmethod
    def from_dict(cls, _dict):
        return cls(**_dict)

class SpintopTestRecordView(object):
    feature_prefix = ('features', lambda feat: feat._type, lambda feat: feat.complete_name)
    data_prefix = () 
    default_view = DefaultPrimitiveView()
    
    def apply(self, record, flatten_dict=True):
        data = OrderedDict()
        
        test_id_data = self.apply_default_test_id(record, flatten_dict=flatten_dict)
        data.update(test_id_data)
        
        for feature in record.features:
            feature_data = self.apply_default_feature(feature, flatten_dict=flatten_dict)
            data.update(feature_data)
            
        return data
    
    def apply_default_feature(self, feature, key_prefix=None, **apply_kwargs):
        if key_prefix is None: key_prefix = self.feature_prefix
        return self.default_view.apply(feature, key_prefix=key_prefix, **apply_kwargs)
    
    def apply_default_test_id(self, record, key_prefix=None, **apply_kwargs):
        if key_prefix is None: key_prefix = self.data_prefix
        data = self.default_view.apply(record.data, key_prefix=key_prefix, **apply_kwargs)
        return data

class SpintopFlatTestRecordBuilder(NamedTuple):
    data: TestRecordSummary
    features: List[FeatureRecord]
    
    def build(self):
        record = SpintopTestRecord(test_record=self.data, features=sorted(self.features, key=lambda feature: feature.index))
        record.compute_stats()
        return record
        
    


import pytest
import os
from spintop.errors import SpintopException
from spintop.auth import AuthModule, AuthException
from spintop.spinhub import SpinHubClientModule

from mock import patch

CREDENTIALS_FILE_MOCK = 'tempcredentials.json'

@pytest.fixture
def auth_module():
    return AuthModule()

@pytest.fixture
def spinhub_module(auth_module, integration_config):
    if not integration_config:
        return None
    else:
        return SpinHubClientModule(url=integration_config.api_url, auth_module=auth_module)

@pytest.mark.integration
def test_login(auth_module, integration_config):
    store.delete_credentials() # make sure none exist
    
    auth_module.login(integration_config.username, integration_config.password)
    
    assert os.path.exists(CREDENTIALS_FILE_MOCK)
    
@pytest.mark.integration
def test_refuse_second_login(auth_module, integration_config):
    assert auth_module.credentials.username == integration_config.username
    
    with pytest.raises(AuthException):
        auth_module.login(integration_config.username, integration_config.password)

@pytest.mark.integration
def test_private_api_endpoint(spinhub_module):
    spinhub_module.test_private_endpoint()
    
@pytest.mark.integration
def test_bad_access_token_causes_refresh(auth_module, spinhub_module):
    auth_module.credentials.access_token = auth_module.credentials.access_token[0:-4]
    spinhub_module.test_private_endpoint()
    
@pytest.mark.integration
def test_bad_access_and_bad_refresh(auth_module, spinhub_module):
    # re-corrupt access
    auth_module.credentials.access_token = auth_module.credentials.access_token[0:-4]
    # corrupt refresh
    auth_module.credentials.refresh_token = auth_module.credentials.refresh_token[0:-4]
    
    with pytest.raises(SpintopException):
        spinhub_module.test_private_endpoint()

@pytest.mark.integration
def test_logout(auth_module):
    auth_module.logout()
    assert not os.path.exists(CREDENTIALS_FILE_MOCK)
    
    
    



import itertools
from dataclasses import dataclass, fields

from .internal import FeatureRecordTemplate
from .collection import SpintopTestRecord

def default_feature_filter(feature):
    return True

class TestRecordTemplate(object):
    def __init__(self, test_records, feature_filter=None):
        self.template_record = SpintopTestRecord()
        self.names_lookup = {}
        self.tests_contained = {}
        
        self.add_test_records(test_records, feature_filter=feature_filter)
        
        
    def add_test_records(self, test_records, feature_filter=None):
        
        for record in test_records:
            self._add_test_record(record, feature_filter=feature_filter)
            
        self.clean()
        
    def _add_test_record(self, test_record, feature_filter=None):
        
        if feature_filter is None: feature_filter = default_feature_filter
        
        features = tuple(feature for feature in test_record.features if feature_filter(feature))
        
        if not features:
            raise ValueError("Filter yields empty features list.")
        
        new_features = tuple(f.copy(as_cls=FeatureRecordTemplate, set_attributes=dict(source=f)) for f in features if (f.name not in self.names_lookup))
        
        self.template_record.features += new_features
        
        for template in new_features:
            self.names_lookup[template.name] = template
        
        # Update stats in template features
        for feature in features:
            self.names_lookup[feature.name].increment_records_count()
            
        self.tests_contained[test_record] = test_record
        
        
    def clean(self):
        # TODO Topological sort before reindexing
        self.template_record.reindex()
        
    def compute_stats(self):
        self.compute_match_scores()
        
    def compute_match_scores(self):
        abs_max_score = len(self.tests_contained)*len(self.names_lookup)
        max_score = 0
        for _, test_record in self.tests_contained.items():
            
            score_int = 0
            for feature in test_record.features:
                # Non-original features meant they were filled in or otherwise not part of the original test
                if feature.original and feature.name in self.names_lookup:
                    # Add score only if original feature and if present in lookup.
                    # If it's not present, it was probably filtered out during creation.
                    score_int += self.names_lookup[feature.name].records_count
            
            max_score = max(max_score, score_int)
            
            test_record.data.template_match_score = score_int
        
        # Normalize all scores based on max score
        for _, test_record in self.tests_contained.items():
            test_record.data.template_match_score_normalized = test_record.data.template_match_score/max_score
    
        
        
    def fill_missing_features(self, test_record):
        """ Based on this template, fill missing features with no data into test_record.
        """
        
        # First step is to reindex test_record features to match the template indices.
        # The names_lookup lookup table allows to quickly find indices.
        for feature in test_record.features:
            feature.index = self.names_lookup[feature.name].index
            
        test_record.fill_missing_from_source(
            self.template_record.features, 
            on_fill=lambda template: template.source.copy(
                with_data=False, 
                set_attributes=dict(original=False)
            )
        )
        


from .collection import (
    SpintopTestRecordCollection, 
    SpintopFlatTestRecordBuilder,
    SpintopTestRecord
)

from .internal import (
    is_type_of,
    BaseDataClass,
    TestIDRecord,
    TestRecordSummary, 
    FeatureRecord,
    MeasureFeatureRecord,
    PhaseFeatureRecord
)

from .queries import Query, multi_query_deserialize, multi_query_serialize

from .serialization import get_serializer, get_bson_serializer, get_json_serializer

from .template import TestRecordTemplate

from .tree_struct import SpintopTreeTestRecord, SpintopTreeTestRecordBuilder



from copy import copy
from collections import Mapping

from anytree import NodeMixin, RenderTree, PreOrderIter, LevelOrderGroupIter
from anytree.render import ContStyle

from .collection import SpintopTestRecord, SpintopFlatTestRecordBuilder
from .internal import TestIDRecord, OutcomeData, PhaseFeatureRecord, MeasureFeatureRecord, TestRecordSummary, NAME_SEPARATOR
    
class SpintopTestRecordError(Exception): pass
    
class AbstractSpintopTestRecordNode(NodeMixin):
    internal_feature = None
    separator = NAME_SEPARATOR
    
    def __init__(self, name, parent=None, **data):
        self._test_id = None
        
        self.name = name
        self.parent = parent
        self.data = {}
        if data:
            self.define_data(**data)
        
    @property
    def tree_name(self):
        return self.separator.join(p.name for p in self.path)
    
    @property
    def test_id(self):
        return self._test_id
    
    def _set_test_id(self, value):
        # Recursively sets the test id of the whole tree.
        self._test_id = value
        for child in self.children:
            child._set_test_id(value)
            
    def _post_attach(self, parent):
        self._test_id = parent._test_id

    def child_tree_name(self, child_name):
        return self.separator.join([self.tree_name, child_name])
    
    def define_id(self, 
            start_datetime,
            dut_id,
            testbench_name):
        
        test_id = TestIDRecord.create(
            testbench_name,
            dut_id,
            start_datetime
        )
        
        # Find the root, which will recursively set the id of all nodes in the tree.
        self.root._set_test_id(test_id)
        
        
    def define_data(self, **data):
        self._update_data(**data)
        
    def _update_data(self, **data):
        self.data.update(data)
        
    def __repr__(self):
        return "%s(name='%s')" % (self.__class__.__name__, self.name)
    
    def __copy__(self):
        return self.__class__(self.name, parent=None, **self.data)
    
    def to_internal(self, index=0):
        if self.test_id is None:
            raise Exception('The test id must be defined before calling')
        internal_cls = self.internal_feature
        obj = internal_cls(
            oid=None, # New class, not bound to any stored one.
            name=self.name,
            version=0,
            depth=self.depth,
            index=index,
            ancestors=[node.name for node in self.path[:-1]],
            original=True,
            test_id=self.test_id,
            user_data={},
            **self._data_for_internal()
        )
        return obj
        
    def _data_for_internal(self):
        return self.data
        
class Phase(AbstractSpintopTestRecordNode):
    internal_feature = PhaseFeatureRecord
    def define_data(self, outcome, duration):
        super(Phase, self).define_data(
            outcome=make_outcome(outcome), 
            duration=duration
        )
    
class Measure(AbstractSpintopTestRecordNode):
    internal_feature = MeasureFeatureRecord
    def define_data(self, outcome, value):
        super(Measure, self).define_data(
            outcome=make_outcome(outcome), 
            value=value
        )
        
class PhasesView(Phase):
    """ Allows to group multiple phases into one. Outcomes are boolean AND and durations are summed.
    """
    phases = []
    
    def define_data(self, wrapped_phases):
        outcome = all([phase.data.get('outcome', True) for phase in wrapped_phases])
        duration = sum([phase.data.get('duration', 0) for phase in wrapped_phases])
        super(PhasesView, self).define_data(outcome=outcome, duration=duration)
        
class SpintopTreeTestRecord(Phase):
    internal_feature = TestRecordSummary
    def __init__(self):
        super(SpintopTreeTestRecord, self).__init__('', parent=None)
    
    def _data_for_internal(self):
        return dict(
            **self.data
        )
    
    def flatten_data(self):
        builder = SpintopFlatTestRecordBuilder(
            data = self.to_internal(),
            features = [node.to_internal(index-1) for index, node in enumerate(PreOrderIter(self)) if node is not self]
        )
        return builder.build()
    
    def pprint(self):
        print(RenderTree(self, style=ContStyle()))
        
    def group_outcomes(self, depth=1):
        for nodes in LevelOrderGroupIter(self, maxlevel=depth):
            if nodes[0].depth < depth-1:
                continue
            else:
                current_group = []
                for node in nodes:
                    if node.data.get('outcome', True):
                        current_group.append(node)
                    else:
                        if current_group:
                            yield current_group, True
                            current_group = []
                    
                        yield [node], False
        
                if current_group:
                    yield current_group, True
                
        

class SpintopTestRecordBuilderContext(object):
    def __init__(self, ensure_unique_names=True):
        self._should_ensure_unique_names = ensure_unique_names
        self._duplicate_key_index = {}
        self._unique_ensured_nodes = {}
        
    def ensure_unique_key(self, key, parent):
        unique_name = parent.child_tree_name(key)
        if self._should_ensure_unique_names and unique_name in self._unique_ensured_nodes:
            suffix = self._duplicate_key_index.get(unique_name, 0)
            self._duplicate_key_index[unique_name] = suffix + 1
            key = key + str(suffix)
            return self.ensure_unique_key(key, parent)
        return key
    
    def add_unique_node(self, node):
        self._unique_ensured_nodes[node.tree_name] = node
    
class SpintopTreeTestRecordBuilder(object):
    def __init__(self, node=None, context=None, **context_kwargs):
        if node is None:
            node = SpintopTreeTestRecord()
        if context is None:
            context = SpintopTestRecordBuilderContext(**context_kwargs)
            
        self.node = node
        self.context = context
    
    def __enter__(self):
        if self.node.parent is None:
            raise SpintopTestRecordError('Cannot use the top level node as a context manager')
        return self
    
    def __exit__(self, *args, **kwargs):
        pass
    
    @property
    def test_record(self):
        return self.node.root
    
    def build(self):
        return self.test_record.flatten_data()
    
    def set_top_level_information(self, 
            start_datetime,
            dut_id,
            testbench_name,
            outcome=None,
            duration=None):
        
        top_level_node = self.test_record
        
        top_level_node.define_id(
            start_datetime,
            dut_id,
            testbench_name
        )
        
        top_level_node.define_data(
            outcome=outcome,
            duration=duration
        )
        
    def new_phase(self, name, outcome, duration):
        return self._new_node(Phase, name,
            outcome=outcome,
            duration=duration
        )
    
    def new_measure(self, name, outcome, value):
        return self._new_node(Measure, name,
            outcome=outcome,
            value=value
        )
    
    def _new_node(self, cls, name, **data):
        name = self.context.ensure_unique_key(name, parent=self.node)
        node = cls(name, parent=self.node, **data)
        self.context.add_unique_node(node)
        return self._wrap(node)
    
    def _wrap(self, node):
        return self.__class__(node, self.context)
    

def make_outcome(outcome):
    if isinstance(outcome, bool):
        outcome = OutcomeData(is_pass=outcome)
    elif isinstance(outcome, Mapping):
        outcome = OutcomeData(**outcome)
    else:
        raise TypeError((
'Outcome (%s) should either be a boolean that indicates if this test passed or failed, ' 
'or a dictionnary of one or more of the following attributes: %s'
        ) % (outcome, [f.name for f in OutcomeData.get_fields()]))
    return outcome


from marshmallow import Schema, fields, pprint


# class SpinhubProfile(Schema):
#     name = fields.String()
#     machine_id = fields.String(allow_none=True)
#     org_id = fields.String(allow_none=True)
#     spinhub_url = fields.String()
    
# class SpintopMachineConfig(Schema):
#     default_profile = fields.String(default='default')
#     profiles = fields.Nested(SpinhubProfile(), many=True)
    
    
from marshmallow import Schema, fields, pprint
from uuid import getnode

from ..auth.schemas import CredentialsSchema

class SpinhubProfile(Schema):
    credentials_key = fields.String(allow_none=True)
    spinhub_url = fields.String(allow_none=True)
    
class SpintopMachineConfig(Schema):
    """
    hardware_uuid: xxxxxxxx
    credentials:
        1234:
            access_token:
            refresh_token:
            username:
    
    """
    default_profile = fields.String(default='default')
    hardware_uuid = fields.String(default=getnode)
    profiles = fields.Dict(keys=fields.Str(), values=fields.Nested(SpinhubProfile()))
    credentials = fields.Dict(keys=fields.Str(), values=fields.Nested(CredentialsSchema()))
    


from unittest.mock import MagicMock
import pytest

from datetime import datetime, timedelta

from spintop.models.internal import DutIDRecord, NO_VERSION
from spintop.transforms import TransformBuilder, DutOp, DutOpLog, join_transforms, transformer

V1_DATETIME = datetime(year=2019, month=1, day=1)
V2_DATETIME = datetime(year=2019, month=2, day=1)

def string_format(datetime):
    return datetime.strftime("%Y-%m-%d %H:%M:%S")

def after(datetime):
    return datetime + timedelta(days=1)

def before(datetime):
    return datetime - timedelta(days=1)

@pytest.fixture()
def op_log():
    log = DutOpLog()
    op1 = DutOp.create(dut_match='my_unit', dut_after={'version': 'v1'}, op_datetime=V1_DATETIME)
    op2 = DutOp.create(dut_match='my_unit', dut_after={'version': 'v2'}, op_datetime=V2_DATETIME)
    
    log.add_op(op2)
    log.add_op(op1)
    return log

def basic_test_on(op_log):
    dut = DutIDRecord.create('my_unit')
    
    assert op_log.get_latest_dut(dut, before(V1_DATETIME)).version == NO_VERSION
    assert op_log.get_latest_dut(dut, after(V1_DATETIME)).version == 'v1'
    assert op_log.get_latest_dut(dut, after(V2_DATETIME)).version == 'v2'

def test_op_log(op_log):
    basic_test_on(op_log)
    
    
def test_op_log_from_dict_ops_by_dut_id_form():
    op_log_dict = dict(
        ops_by_dut_id=dict(
            my_unit=[
                dict(
                    dut_after={'version': 'v1'},
                    op_datetime=string_format(V1_DATETIME)
                ),
                dict(
                    dut_after={'version': 'v2'},
                    op_datetime=string_format(V2_DATETIME)
                )
            ]
        )
    )
    op_log = DutOpLog.from_dict(op_log_dict)
    basic_test_on(op_log)
    
def test_op_log_from_dict_ops_form():
    op_log_dict = dict(
        ops=[
            dict(
                dut_match='my_unit',
                dut_after={'version': 'v1'},
                op_datetime=string_format(V1_DATETIME)
            ),
            dict(
                dut_match='my_unit',
                dut_after={'version': 'v2'},
                op_datetime=string_format(V2_DATETIME)
            )
        ]
    )
    op_log = DutOpLog.from_dict(op_log_dict)
    basic_test_on(op_log)
    
def test_op_log_from_dict_multi_match():
    op_log_dict = dict(
        ops=[
            dict(
                dut_match=['my_unit', 'other_unit'],
                dut_after={'version': 'v1'},
                op_datetime=string_format(V1_DATETIME)
            ),
            dict(
                dut_match='my_unit',
                dut_after={'version': 'v2'},
                op_datetime=string_format(V2_DATETIME)
            )
        ]
    )
    op_log = DutOpLog.from_dict(op_log_dict)
    basic_test_on(op_log)
    
    
def append_arg_transform_factory(arg):
    @transformer
    def _transform(obj):
        return obj + arg
    return _transform

def test_transform_builder_steps():
    builder = TransformBuilder(
        append_arg_transform_factory('a'),
        append_arg_transform_factory('b'),
        append_arg_transform_factory('c')
    )

    assert builder("") == "abc"
    
def test_transform_builder_append():
    builder = TransformBuilder(
        append_arg_transform_factory('a')
    )
    
    builder.append(append_arg_transform_factory('b'))
    
    assert builder("") == "ab"
    
def test_transform_builder_prepend():
    builder = TransformBuilder(
        append_arg_transform_factory('b')
    )
    
    builder.prepend(append_arg_transform_factory('a'))
    
    assert builder("") == "ab"
    
def test_transform_builder_append_prepend():
    builder = TransformBuilder(
        append_arg_transform_factory('b')
    )
    
    builder.prepend(append_arg_transform_factory('a'))
    builder.append(append_arg_transform_factory('c'))
    
    assert builder("") == "abc"
    
def test_transform_builder_split():
    
    b_transform = MagicMock(side_effect=lambda obj: obj + "b")
    
    builder = TransformBuilder(b_transform)
    
    builder_a, builder_c = builder.split(2)
    
    builder_a.prepend(append_arg_transform_factory('a'))
    
    builder_c.append(append_arg_transform_factory('c'))
    
    assert builder_a("") == "ab"
    assert builder_c("") == "bc"
    
    assert b_transform.call_count == 2, "The shared transform should be called twice"
    
def test_transform_builder_append_factories():
    ab_builder = TransformBuilder(
        append_arg_transform_factory('a'),
        append_arg_transform_factory('b')
    )
    
    c_builder = TransformBuilder(
        append_arg_transform_factory('c')
    )
    
    ab_builder.append(c_builder)
    
    assert ab_builder("") == "abc"
    
def test_transform_builder_prepend_factories():
    bc_builder = TransformBuilder(
        append_arg_transform_factory('b'),
        append_arg_transform_factory('c')
    )
    
    a_builder = TransformBuilder(
        append_arg_transform_factory('a')
    )
    
    bc_builder.prepend(a_builder)
    
    assert bc_builder("") == "abc"
    
def test_transform_builder_branch_in_and_out():
    """ 
    +A -> +B \           / +D
              |-> +C -> |
    +C -> +C /           \ +E
    
    Results should be
    - ABCD
    - ABCE
    - CCCD
    - CCCE
    """ 
    
    
    ab = join_transforms(
        append_arg_transform_factory('a'),
        append_arg_transform_factory('b')
    )
    
    cc = join_transforms(
        append_arg_transform_factory('c'),
        append_arg_transform_factory('c')
    )
    
    c = append_arg_transform_factory('c')
    
    d = append_arg_transform_factory('d')
    
    e = append_arg_transform_factory('e')
    
    abcd = ab + c + d
    abce = ab + c + e
    cccd = cc + c + d
    ccce = cc + c + e
    
    assert abcd("") == "abcd"
    assert abce("") == "abce"
    assert cccd("") == "cccd"
    assert ccce("") == "ccce"

@transformer
def append_d_and_e(obj):
    for append in ('d', 'e'):
        yield obj + append

def test_transform_generator():
    """ 
           / +D
    +C -> |
           \ +E
    """

    c = append_arg_transform_factory('c')

    
    cde = c + append_d_and_e
    result = cde.generate("")
    assert list(result) == ['cd', 'ce']


def test_transform_generator_alone():

    result = append_d_and_e.generate("")
    assert list(result) == ['d', 'e']


import dateutil.parser

from collections import defaultdict, Sequence

from spintop.models import (
    SpintopTestRecordCollection
)
from spintop.models.internal import DutOp

from .base import Transformer

from ..utils import load_yaml_file
from ..logs import _logger

logger = _logger('transforms.op_log')

def and_(*transformers):
    """Record transformer that calls all transformers in order."""
    def call_all_transformers(value):
        for transformer in transformers:
            value = transformer(value)
        return value
    return call_all_transformers
        
class DutOpLog(object):
    def __init__(self):
        self.dut_ops = defaultdict(list)
    
    @classmethod
    def from_yaml(cls, yaml_filename):
        """
        """
        content = load_yaml_file(yaml_filename)
        return cls.from_dict(content)
    
    @classmethod
    def from_dict(cls, content):
        """ Can contain two keys:
        
        ops_by_dut_id: A dut_id indexed list of operations
        ops: A list of operation
        
        (in yaml format)
        ops:
          - dut_match: 'my_unit'
            dut_after: 
                id: 'my_unit' # optional
                version: 'my_version'
            op_datetime: '2018-02-11 11:02:57'
          - (other ops...)
          
        OR
        
        ops_by_dut_id:
            my_unit:
                - dut_after:
                    version: 'my_version'
                  op_datetime: '2018-02-11 11:02:57'
        """
        
        ops = content.get('ops', None)
        ops_by_dut_id = content.get('ops_by_dut_id', None)
        if ops:
            return cls._parse_ops_array(ops)
        elif ops_by_dut_id:
            return cls._parse_ops_by_dut_id_dict(ops_by_dut_id)
        else:
            raise ValueError('Dict content must have either a ops or ops_by_dut_id top-level key. Received keys: {}'.format(
                list(content.keys())
            ))
        
    
    @classmethod
    def _parse_ops_by_dut_id_dict(cls, ops_dict):
        all_ops = []
        for dut_match, ops in ops_dict.items():
            for op in ops:
                if 'dut_match' not in op:
                    op['dut_match'] = dut_match
            all_ops += ops
        
        return cls._parse_ops_array(all_ops)
        
    
    @classmethod
    def _parse_ops_array(cls, ops_list):
        op_log = cls()
        for op in ops_list:
            datetime_str = op['op_datetime']
            dut_matches = op['dut_match']
            
            if not isinstance(dut_matches, Sequence) or isinstance(dut_matches, str):
                dut_matches = [dut_matches]
            
            dut_after = op['dut_after']
            
            for dut_match in dut_matches:
                op = DutOp.create(dut_match, dut_after, dateutil.parser.parse(datetime_str))
                op_log.add_op(op)
        return op_log
    
    def add_op(self, dut_op):
        key = dut_op.dut_match.id
        self.dut_ops[key].append(dut_op)
        self.dut_ops[key].sort(key=lambda op: op.op_datetime, reverse=True)
        
    def get_latest_dut(self, dut, on_datetime):
        ops = self.dut_ops[dut.id]
        for op in ops:
            if op.does_applies_to(dut, on_datetime):
                return op.apply(dut)
        # No op match
        return dut
        

class DutOpLogTransformer(Transformer):
    """Record transformer that attaches version information to DUTs
    based on an operation log."""
    def __init__(self, op_log):
        self.op_log = op_log
        
    def __call__(self, test_record):
        on_datetime = test_record.test_id.start_datetime
        dut = test_record.test_id.dut
        
        new_dut = self.op_log.get_latest_dut(dut, on_datetime)
        
        if new_dut != test_record.test_id.dut:
            logger.info('Applied operation on {} @ {} to become {}'.format(
                test_record.test_id.dut,
                test_record.test_id.start_datetime,
                new_dut
            ))
        else:
            logger.debug('No operation applied on {} @ {}'.format(
                test_record.test_id.dut,
                test_record.test_id.start_datetime
            ))
        
        test_record.test_id.dut = new_dut
        return test_record


        


""" Storage of the access token and refresh token to allow long lived authentication sessions.
"""
import yaml
import os
from pprint import pformat

from ..storage import SITE_DATA_DIR
from .schemas import SpintopMachineConfig

from ..logs import _logger

logger = _logger('config-storage')

class FileConfigStorageProvider(object):
    def __init__(self, config_filepath):
        self.config_filepath = config_filepath
    
    def store_spintop_config(self, config):
        data = SpintopMachineConfig().dump(config)
        logger.info('Saving config')
        logger.debug('Config content:\n' + pformat(data))
        with open(self.config_filepath, 'w+') as machine_file:
            yaml.dump(data, machine_file)
        return data

    def retrieve_spintop_config(self):
        if os.path.exists(self.config_filepath):
            logger.debug('Loading config content')
            with open(self.config_filepath) as machine_file:
                data = yaml.safe_load(machine_file)
                return SpintopMachineConfig().load(data)
        else:
            logger.info('No config file to load')
            return None
        
    def delete_spintop_config(self):
        if os.path.exists(self.config_filepath):
            os.remove(self.config_filepath)
        



from contextlib import contextmanager
import cProfile, pstats, io

import yaml

def load_yaml_file(filename):
    with open(filename,'r') as readfile:
        data = readfile.read()
        return yaml.safe_load(data)
    

@contextmanager
def profiling(output_filename):
    pr = cProfile.Profile()
    pr.enable()
    try:
        yield
    finally:
        pr.disable()
        s = io.StringIO()
        sortby = 'cumulative'
        with open(output_filename, 'w+') as profile_stream:
            ps = pstats.Stats(pr, stream=profile_stream).sort_stats(sortby)
            ps.print_stats()


import inspect
from collections import Mapping

from dataclasses import dataclass, fields, _MISSING_TYPE, MISSING, field, asdict

from .view import DataClassPrimitiveView, join_key_prefix

_TYPE_MAP = {}

def register_type(_type):
    def _wrap_cls(cls):
        cls._type = _type
        assert _type not in _TYPE_MAP, "Type %s is already registered to class %s" % (_type, _TYPE_MAP[_type])
        _TYPE_MAP[_type] = cls
        return cls
    return _wrap_cls

class CacheByCls():
    
    def __init__(self):
        self.value_by_cls = {}
        
    def retrieve(self, cls):
        if not self.value_by_cls.get(cls, None):
            self.value_by_cls[cls] = self._get(cls)
        return self.value_by_cls[cls]
    
    def _get(self, cls):
        raise NotImplementedError()
    
class FieldsCache(CacheByCls):
    """The fields() function is quite slow. This allows to store its 
    return value per cls, as it shouldn't change dynamically during
    runtime."""
    
    def _get(self, cls):
        return fields(cls)
    
class DefaultViewsCache(CacheByCls):
    def _get(self, cls):
        return DataClassPrimitiveView(
            cls.get_default_view_mapping()
        )

_fields_cache = FieldsCache()
_default_views_cache = DefaultViewsCache()

class KeyValueOf(object):
    def __init__(self, attr_name):
        self.attr_name = attr_name
    
    def __call__(self, obj):
        return getattr(obj, self.attr_name)

class MetaBaseDataClass(type):
    def __getattr__(cls, key):
        """ Allows fields to be accessed from class. I.e. 
        
        @dataclass()
        class MyDataClass(BaseDataClass):
            my_field: str
            
        MyDataClass.my_field.name == 'my_field'
        """
        try:
            return cls.get_field(key)
        except AttributeError:
            return super().__getattribute__(key)
    
    def get_field(cls, key, parent_fields=()):
        # no cache here has the cls might be being build.
        for field in fields(cls):
            if field.name == key:
                return MetaField(field, parents=parent_fields)
        else:
            raise AttributeError(key)


class MetaField(object):
    def __init__(self, field, parents=()):
        self.__parents = parents
        self.__field = field
    
    @property
    def name_(self):
        names = tuple(parent.__field.name for parent in (self.__parents + (self,))) 
        return '.'.join(names)

    def __getattr__(self, key):
        try:
            return self.__field.type.get_field(key, parent_fields=self.__parents + (self,))
        except AttributeError:
            return getattr(self.__field, key)
    
    def __eq__(self, other):
        if type(other) is type(self):
            return self.__members() == other.__members()
        else:
            return False

    def __hash__(self):
        return hash(self.__field)
    
    def __eq__(self, other):
        return self.__field == other.__field

    
# def make_operation(field, op_on_value):
#     return lambda real_obj: op_on_value(getattr(real_obj, field.name))

def model_dataclass(**kwargs):
    return dataclass(init=False, **kwargs)

@model_dataclass()
@register_type('_base')
class BaseDataClass(metaclass=MetaBaseDataClass):
    
    def __init__(self, *args, **kwargs):
        if args:
            raise NotImplementedError('Positionnal args not supported with BaseDataClass descendants.')
        defaults = self.defaults()
        defaults.update(kwargs)
        
        self.__dict__.update(defaults)
        
    @classmethod
    def defaults(cls, **others):
        return others
    
    @classmethod
    def null(cls, **not_null_fields):
        """ Creates an instance with all attributes set to None. 
        Calls null() on any sub- BaseDataClass fields.
        """
        arguments = {}
        fields = cls.get_fields()
        for field in fields:
            if inspect.isclass(field.type) and issubclass(field.type, BaseDataClass):
                value = field.type.null()
            elif not isinstance(field.default_factory, _MISSING_TYPE):
                value = field.default_factory()
            elif field.default != MISSING:
                value = field.default
            else:
                value = None
                # try:
                #     value = field.type()
                # except TypeError:
                #     raise Exception('Field {} of type {} from {} does not support call without arguments for null()'.format(field.name, field.type, cls))
            arguments[field.name] = value
        
        arguments.update(not_null_fields)
        return cls(**arguments)
    
    @classmethod
    def get_fields(cls):
        return _fields_cache.retrieve(cls)
    
    @classmethod
    def get_data_fields(cls):
        """ Sub classed by FeatureRecord below.
        """
        return cls.get_fields()
        
    @classmethod
    def get_default_view(cls):
        return _default_views_cache.retrieve(cls)
        
    @classmethod
    def get_default_view_mapping(cls, with_key_prefix=None):
        if with_key_prefix is not None:
            key_creator = lambda key: join_key_prefix(with_key_prefix, key)
        else:
            key_creator = lambda key: key
        
        # The default view uses the fields defined by the get_data_fields class method.
        fields = cls.get_data_fields()
        
        return {key_creator(field.name): field for field in fields}
        
    
    def copy(self, with_data=True, as_cls=None, set_attributes={}):
        if with_data:
            nullify_fields = []
        else:
            nullify_fields = self.get_data_fields()
            
        copy_obj = copy_nullify_fields(self, nullify_fields, target_cls=as_cls)
        for name, value in set_attributes.items():
            setattr(copy_obj, name, value)
        return copy_obj
    
    def as_dict(self):
        """ Transforms this object recursively into a dict. Adds the static
        _type field to the data"""
        data = asdict(self)
        data.update({'_type': self._type})
        return data
    
    @staticmethod
    def cls_data_from_dict(cls, data):
        _type = data.pop('_type', None)
        if _type:
            cls = _TYPE_MAP[_type]
        return cls, data
    
    def get_recursive(self, as_str=None, as_array=None):
        if as_str:
            as_array = as_str.split('.')
        
        this_attr_name = as_array.pop(0)
        this_attr = getattr(self, this_attr_name)
        if as_array:
            return this_attr.get_recursive(as_array=as_array)
        else:
            return this_attr


def copy_nullify_fields(obj, null_fields, target_cls=None):
    if target_cls is None:
        target_cls = obj.__class__
    
    copy_obj = target_cls.null()
    not_null_fields = (set(obj.get_fields()) - set(null_fields)) & set(target_cls.get_fields())
    for field in not_null_fields:
        setattr(copy_obj, field.name, getattr(obj, field.name))
    
    return copy_obj



import os

import requests 
import platform
from pprint import pformat
from json.decoder import JSONDecodeError
from uuid import uuid4

from functools import partial
from ..errors import SpintopException

from ..auth import AuthModule

from ..logs import _logger
from ..compat import VERSION
from ..storage import TEMP_DIR

logger = _logger('spinhub-client')

TEMPORARY_DOWNLOAD_FOLDER = os.path.join(TEMP_DIR, '.spinhub-download-cache')
if not os.path.exists(TEMPORARY_DOWNLOAD_FOLDER):
    os.makedirs(TEMPORARY_DOWNLOAD_FOLDER)
    
class SpinHubClientError(SpintopException):
    pass

class SpinHubAPIError(SpintopException):
    def __init__(self, resp):
        try:
            json = resp.json()
        except JSONDecodeError:
            json = {}
        
        error_type = json.get('error_type', 'UNKNOWN')
        error_messages = json.get('error_messages', {})
        messages_as_string = '\n\t'.join(['%s: %s' % (key, value) for key, value in error_messages.items()])
        
        if not messages_as_string:
            messages_as_string = json.get('error_message') # singular, the direct exception message
            
        message = 'A {status} error of type "{error_type}" occured while contacting API.\n\t{messages}'.format(
            status=resp.status_code, 
            error_type=error_type,
            messages=messages_as_string
        )
        super(SpinHubAPIError, self).__init__(message)
        self.resp = resp

# def SpinHubClient(spinhub_url=None, auth=None, org=None):
#     if auth is None:
#         auth = AuthModule()
#     return SpinHubAPIModule(spinhub_url, auth, org=org)

def SpinHubSession(prefix, auth_module):
    if prefix is None:                                                                                                                                                                                                                                                                                                                             
        prefix = ""                                                                                                                                                                                                                                                                                                                                
    else:                                                                                                                                                                                                                                                                                                                                          
        prefix = prefix.rstrip('/')                                                                                                                                                                                                                                                                                                      

    def attempt_request(f, method, url, *args, **kwargs):
        if 'headers' not in kwargs: kwargs['headers'] = {}
        kwargs['headers'].update(auth_module.get_auth_headers())
        resp = f(method, prefix + url, *args, **kwargs)           
        if resp.status_code >= 400:
            raise SpinHubAPIError(resp)
        return resp
    
    def new_request(prefix, *args, **kwargs):
        try:
            resp = attempt_request(*args, **kwargs)  
        except SpinHubAPIError as e:
            if e.resp.status_code == 401:
                # Attempt ONCE to refresh the credentials and retry.
                auth_module.refresh_credentials()
                resp = attempt_request(*args, **kwargs)
            else:
                raise
        return resp
    
    s = requests.Session()                                                                                                                                                                                                                                                                                                                     
    s.request = partial(new_request, prefix, s.request)                                                                                                                                                                                                                                                                                            
    return s  

class SpinHubClientModule(object):
    def __init__(self, config, auth):
        self.config = config
        
        self._session = None
        
        self.auth = auth
        
        self.default_org_id = None
        
        self._user_info = None
        self._default_org = None
        self._machine = None
    
    @property
    def session(self):
        if not self._session:
            profile = self.config.get_selected_profile()
            url = profile['spinhub_url']
            self.default_org_id = profile.get('org_id')
            self._session = SpinHubSession(url, self.auth)
        return self._session
    
    def test_private_endpoint(self):
        resp = self.session.get('/private')
        return resp
    
    @property
    def user_info(self):
        if self._user_info is None:
            self._user_info = self.session.get('/userinfo').json()
        return self._user_info
    
    @property
    def default_org(self):
        if self._default_org is None:
            possible_orgs = self.user_info['organizations']
            selected_org = None
            if self.default_org_id:
                for org in possible_orgs:
                    if org['uuid'].startswith(self.default_org_id):
                        selected_org = org
                        break
                else:
                    raise SpinHubClientError('Unable to find org with ID %s associated with this machine.' % self.default_org_id)
            else:
                selected_org = possible_orgs[0]
                logger.info('No org specified, using the default defined by SpinHub for this user: %s' % selected_org['key'])
            self._default_org = selected_org
        
        return self._default_org
    
    def get_this_machine_id(self):
        return self.config.get_selected_profile()['machine_id']
    
    def get_token(self):
        token_url = self._get_org_urls('token')
        return self.session.get(token_url).text
    
    @property
    def machine(self):
        if self._machine is None:
            machines_url = self._get_org_urls('machines')
            this_machine_id = self.get_this_machine_id()
            logger.info('Local machine ID is %s' % this_machine_id)
            self._machine = self.session.get(machines_url + '/' + this_machine_id).json()
        return self._machine
    

    def create_task(self, task_type, user_id=None):
        if user_id is None:
            user_id = self.get_this_machine_id()
        machines_url = self._get_org_urls('machines')
        self.session.post(machines_url + '/' + user_id + '/tasks', json=dict(
            task_type=task_type
        ))
    
    def task_acknowledged(self, task_id):
        return self.update_task_status(task_id, 'ACKNOWLEDGED')
    
    def task_started(self, task_id):
        return self.update_task_status(task_id, 'IN_PROGRESS')
    
    def task_done(self, task_id, success=True):
        return self.update_task_status(task_id, 'DONE', task_success=success)
        
    def update_task_status(self, task_id, task_status, task_success=None):
        user_id = self.get_this_machine_id()
        machines_url = self._get_org_urls('machines')
        self.session.put(machines_url + '/' + user_id + '/tasks', json=dict(
            uuid=task_id,
            task_status=task_status,
            task_success=task_success
        ))
        
    
    def get_tasks(self, **query_params):
        machines_url = self._get_org_urls('machines')
        this_machine_id = self.get_this_machine_id()
        return self.session.get(machines_url + '/' + this_machine_id + '/tasks', params=query_params).json()
    
    def _get_org_urls(self, sub_key):
        return self.default_org['_links'][sub_key]
    
    def register_machine(self, name, token, org_id=None):
        if org_id:
            machines_url = '/orgs/%s/machines' % org_id
        else:
            machines_url = self._get_org_urls('machines')
            
        logger.info('Registering machine')
        
        params = {}
        if token:
            params.update({'secret_code': token})
        
        response = self.session.put(machines_url + '/register', json=dict(
            nickname= name,
            hostname= platform.node(),
            client_version=VERSION
        ), params=params)
        
        data = response.json()
        self.config.update_profile(machine_id= data['uuid'])
        self.config.save()
    
    def upload_file(self, object_key, version, filepath, checksums={}):
        objects_url = self._get_org_urls('objects')
        logger.info('Asking SpinHub to upload new file %s @ %s' % (filepath, object_key))
        # Generate a new URL for upload.
        
        _, filename = os.path.split(filepath)
        response = self.session.put(objects_url, json=dict(
            object_key=object_key,
            version_name=version,
            filename=filename,
            checksums=checksums
        ))
        response_data = response.json()
        upload_uri = response_data['upload_uri']
        upload_fields = response_data['upload_fields']
        
        logger.info('Beginning upload at SpinHub provided URI')
        with open(filepath, 'rb') as file_to_upload:
            files = {'file': file_to_upload}
            resp = requests.post(upload_uri, files=files, data=upload_fields)
        
        self.confirm_uploaded_file(response_data)
        return response_data
    
    def confirm_uploaded_file(self, upload_response):
        confirm_uri = self.get_data_blob_uri(upload_response['data_blob'], key='confirm')
        self.session.put(confirm_uri)
    
    def download_uploaded_file(self, upload_response, stream):
        download_uri = self.get_data_blob_uri(upload_response['data_blob'])
        return self.download_data_blob(download_uri, stream)
    
    def get_data_blob_uri(self, data_blob, key='download'):
        return data_blob['_links'][key]
    
    def get_tag_download_uri(self, tag):
        return tag['_links']['download']
    
    def download_file(self, object_key, target_folder=TEMPORARY_DOWNLOAD_FOLDER, index=0):
        objects_url = self._get_org_urls('objects')
        response = self.session.get(objects_url + '/' + object_key)
        object_details = response.json()
        tag = object_details['tags'][index]
        
        logger.debug(pformat(tag))
        logger.info('Downloading {object_key} [version={tagname}][{index}]'.format(
            object_key=object_key,
            tagname=tag['tagname'],
            index=index
        ))
        download_uri = self.get_tag_download_uri(tag)
        
        return self.download_data_blob(download_uri, target_folder=target_folder)
        
    def download_data_blob(self, download_uri, target_folder=None, stream=None):
        data_object = self.session.get(download_uri).json()
        url = data_object['download_uri']
        filename = data_object['filename']
        resp = requests.get(url, allow_redirects=True)
        
        if stream:
            stream.write(resp.content)
            return None
        else:
            filepath = os.path.join(target_folder, filename)
            with open(filepath, 'wb+') as writefile:
                writefile.write(resp.content)
            return filepath
        
        